/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.xmlgraphics.io;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URL;

import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.URIResolver;

/**
 * An adapter between {@link URIResolver} to {@link ResourceResolver}. This adapter allows users
 * to utilize the resolvers from the XML library for resource acquisition.
 */
public class URIResolverAdapter implements ResourceResolver {

    private final URIResolver resolver;

    /**
     * @param resolver the desired {@link URIResolver}
     */
    public URIResolverAdapter(URIResolver resolver) {
        this.resolver = resolver;
    }

    /** {@inheritDoc} */
    public Resource getResource(URI uri) throws IOException {
        try {
            Source src = resolver.resolve(uri.toASCIIString(), null);
            InputStream resourceStream = XmlSourceUtil.getInputStream(src);

            if (resourceStream == null) {
                URL url = new URL(src.getSystemId());
                resourceStream = url.openStream();
            }
            return new Resource(resourceStream);
        } catch (TransformerException e) {
            throw new IOException(e.getMessage());
        }
    }

    /** {@inheritDoc} */
    public OutputStream getOutputStream(URI uri) throws IOException {
        try {
            Source src = resolver.resolve(uri.toASCIIString(), null);
            return new URL(src.getSystemId()).openConnection().getOutputStream();
        } catch (TransformerException te) {
            throw new IOException(te.getMessage());
        }
    }
}
